/**************************************************************************** 
*  Copyright 2008 - 2012 C&H Technologies, Inc.  All Rights Reserved.					    * 
*****************************************************************************/
/****************************************************************************
*
*	m214_ansi.c
*
*	APIS driver for the m214 Precision Voltage Reference m-module  
*	The m214 is manufactured and supported by C&H Technologies, Inc.
*
*	Developed in National Instrument's LabWindows/CVI Version 8.5
*
*	C&H Part Number: 11029269
*
*	Original Release: David Clark    
*
* Author(s): Bobby Schultz
*            David Clark
*						 Gary Guilbeaux
*
*
*	Modification History:
*		Rev 3.1  10/05/12  Also see chm214.c modifications history						GWG
*   				           Increased precision of SetVoltage and GetVoltage		GWG
*                      	functions.
*		Rev 3.2  11/15/12  Corrected a bug in m214_SetVoltage that computed		GWG
*												the wrong voltage bits for exactly 10.0V
*	
*
*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "m214_ansi.h"

/**************************************************************************** 
*  Non Exported Defines
*****************************************************************************/
/*  Driver Revision  */
#define		M214_DRIVER_REV				"3.2"

/*  Instrument ID Defines  */
#define		M214_MODEL_CODE					          0xD6 			// decimal 214
#define		M214_VXI_MANF_ID									0x0FC1
#define		M214_VXI_MODEL_CODE								0x0FDA


/**************************************************************************** 
*  Error Checking Macros
*****************************************************************************/
#ifndef checkParamRange
#define  checkParamRange(value, min, max, param)		if ((value > max) || (value < min)) { \
																											error = M214_ERROR_PARAMETER1 + param-1;\
																											goto Error; \
																										} else 
#endif																				
																				
/*  M214 may generate a warning if a unsigned int is sent to the checkParamRange() macro with a 
		min value of 0.  Use the following checkUnsignedParamRange() to bypass the warning.
*/		
#ifndef checkUnsignedParamRange
#define  checkUnsignedParamRange(value, max, param)		if (value > max) { \
																												error = M214_ERROR_PARAMETER1 + param-1;\
																												goto Error; \
																												} else 
#endif																				


#ifndef checkParamPointer
#define  checkParamPointer(pointer, param)		if (pointer == NULL) { \
																							error = M214_ERROR_PARAMETER1 + param-1;\
																							goto Error; \
																							} else 
#endif																				

																				
#ifndef checkErr
#define checkErr(fCall)      if (error = (fCall), (error = (error < 0) ? error : M214_SUCCESS)) \
                                 {goto Error;}  else
#endif

#ifndef apisCheckErr
#define apisCheckErr(fCall)      if (error = (fCall), (error = (error != 0) ? error : M214_SUCCESS)) \
                                 {	handle->apis_error = error;\
																		error = M214_ERROR_APIS; \
																		goto Error;}  else
#endif

#ifndef apisCheckErrInit
#define apisCheckErrInit(fCall)      if (error = (fCall), (error = (error != 0) ? error : M214_SUCCESS)) \
                                 {	(*handle)->apis_error = error;\
																		error = M214_ERROR_APIS; \
																		goto Error;}  else
#endif

/**************************************************************************** 
*  Constants
*****************************************************************************/
static const char *errorMessageTable[M214_SPECIFIC_ERROR_END - M214_SPECIFIC_ERROR_BASE] = 
						{	"Timeout waiting for calibration access to complete",
							"An invalid parameter was passed to the function in parameter 1",
							"An invalid parameter was passed to the function in parameter 2",
							"An invalid parameter was passed to the function in parameter 3",
							"An invalid parameter was passed to the function in parameter 4",
							"An invalid parameter was passed to the function in parameter 5",
							"An invalid parameter was passed to the function in parameter 6",
							"An invalid parameter was passed to the function in parameter 7",
							"An invalid parameter was passed to the function in parameter 8",
							"An invalid parameter was passed to the function in parameter 9",
							"An invalid parameter was passed to the function in parameter 10",
							"The module returned an invalid ID.",
							"An error occurred in the APIS Platform Support Module.  Use the function ma204_GetApisError() to retrieve the APIS error code.",
							"The operation could not be completed because the system is out of memory."
						};

static const char *warningMessageTable[M214_SPECIFIC_WARNING_END - M214_SPECIFIC_WARNING_BASE] = {""	};


/**************************************************************************** 
*  Non Exported Functions
*****************************************************************************/
static Int32 m214_IDWriteBit(M214_HANDLE handle, Bool value);
static Int32 m214_IDWriteByte(M214_HANDLE handle, UInt8 value);
static Int32 m214_IDReadBit(M214_HANDLE handle, Bool *value);
static Int32 m214_IDReadByte(M214_HANDLE handle, UInt8 *value);


/*****************************************************************************
*	m214_IDWriteBit
*
*	Write a single bit to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m214_IDWriteBit(M214_HANDLE handle, Bool value)
{	
	Int32 error = M214_SUCCESS;

/*  data with clock low  */
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, (0x4 | (value & 0x1))));
/*  data with clock high  */	 
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, (0x6 | (value & 0x1))));

	Error:	
	return error;
}



/*****************************************************************************
*	m214_IDWriteByte
*
*	Write a command byte to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m214_IDWriteByte(M214_HANDLE handle, UInt8 value)
{	
	Int32 error = M214_SUCCESS;
	Int32 i;
	
/*  Initialize  */	
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, 0));
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, 0x4));
	
/*  Write start bit  */	
	checkErr(m214_IDWriteBit(handle, 1));
	for(i=0; i<=7; i++)
	{
	/*  Write data bits  */		
		checkErr(m214_IDWriteBit(handle, ((value &0x80)>>7)));
		value = value << 1;
	}
Error:	
	return error;
}



/*****************************************************************************
*	m214_IDReadBit
*
*	Reads a single bit from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m214_IDReadBit(M214_HANDLE handle, Bool *value)
{	
	Int32 error = M214_SUCCESS;
	UInt16 data;

/*  clock low  */ 	
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, 0x4));
	
/*  clock high  */	
	checkErr(m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, 0x6));
	
/*  Read data  */	
	checkErr(m214_ReadRegister(handle, M214_ID_ACCESS_REGISTER, &data));
	*value = (Bool)data;

Error:	
	return error;
}


/*****************************************************************************
*	m214_IDReadByte
*
*	Reads a byte from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m214_IDReadByte(M214_HANDLE handle, UInt8 *value)
{	
	Int32 error = M214_SUCCESS;
	Int16 i;
	Bool data;
	
	*value = 0;
	for(i=7; i>=0; i--)
	{ 
	/*  read data 1 bit at a time  */	
		checkErr(m214_IDReadBit(handle, &data));
		*value |= ((data & 0x1) << i);
	}
	
Error:	
	return error;
}


/*****************************************************************************
*	ReadID
*
* reads configuration and model number of the M214 module
******************************************************************************/
Int32 m214_ReadID(M214_HANDLE handle, UInt8* configuration, UInt8* model)  
{
	Int32 error = M214_SUCCESS;
	UInt16 data;
	
	checkParamPointer(handle, 1);
	checkParamPointer(configuration, 2);
	checkParamPointer(model, 3);
	
	checkErr(m214_ReadRegister (handle, M214_ID_REG, &data)); 
	*configuration = (data & 0x0FF00) >> M214_CONFIG_BIT_NUM;
	*model = data & 0x0FF;

Error:
	return error;
}


/*****************************************************************************
*	m214_RevisionQuery
*
*	Returns the version number of this driver as well as the hardware version and 
*	firmware version of the module.  If you pass NULL to the handle parameter
*	this function will only return the driver revision.  You can pall a NULL
*	pointer to and of the parameters and the function will not return that 
*	parameter.
*
*	driver_rev	- returned as a string with max size 256
*	fw_rev			-	retured as a binary value
*	hw_rec			-	returned as a binary value
******************************************************************************/
Int32 m214_RevisionQuery(M214_HANDLE handle, Int8 *driver_rev, UInt8 *fw_rev, UInt8 *hw_rev)
{	Int32 error = M214_SUCCESS;
	UInt16 data;

	if(driver_rev != NULL)
		sprintf(driver_rev, "%s", M214_DRIVER_REV);

	if(handle != NULL)
	{	/*  Read the revision register  */
		checkErr(m214_ReadRegister (handle, M214_REVISION_REG, &data));

		if(fw_rev != NULL)
			*fw_rev = (data & 0x0FF00) >> M214_FW_REV_BIT_NUM;
		if(hw_rev != NULL)
			*hw_rev = data & 0x0FF;
	}
	
	
Error:
	return error;
}


/*******************************************************
* GetTemperature
*
* reads the on-board temperature register, calculates
* and returns the temperature in degrees Celcius
********************************************************/
Int32 m214_GetTemperature(M214_HANDLE handle, Real64 *temp_1, Real64 *temp_2)   
{	Int32 error = M214_SUCCESS;
	UInt16 data;

/*  Check Parameters  */
	checkParamPointer(temp_1, 2);
	checkParamPointer(temp_2, 3);

/*  Read the Temperature Register  */
	checkErr(m214_ReadRegister (handle, M214_TEMPERATURE_REG, &data));

/*  Calculate the temperature values from the register reading  */	
	*temp_2 =  (((data >> 8) & 0x00FF) / M214_TEMPERATURE_COEFF) - M214_TEMPERATURE_OFFSET;
	*temp_1 =  ((data & 0x00FF) / M214_TEMPERATURE_COEFF) - M214_TEMPERATURE_OFFSET;

Error:
	return error;
}


/*******************************************************
* GetAnalogInput
*
* reads the analog input voltage from the external analog
* input register
********************************************************/
Int32 m214_GetAnalogInput(M214_HANDLE handle, Real64 *voltage)   
{	Int32 error = M214_SUCCESS;
	UInt16 data;

/*  Check Parameters  */
	checkParamPointer(voltage, 2);

/*  Read the Analog Input Register  */
	checkErr(m214_ReadRegister (handle, M214_ANALOG_INPUT_REG, &data));
	
/*  Caluclate the real voltage from the register reading  */	
	*voltage = data * M214_ANALOG_VOLT_COEFF;
Error:
	return error;
}



/*******************************************************************
* ReadCalibrationValue
*
* reads the current calibration value at the specified 
* calibration address
*********************************************************************/
Int32 m214_ReadCalibrationValue(M214_HANDLE handle, UInt16 address, UInt16 *data)      // BS
{
	Int32 error = M214_SUCCESS;
	UInt16 read;

/*  Check Parameters  */
	checkParamRange(address, 0, 0xFFFF, 2);
	checkParamPointer(data, 3);

/*  Perform the Read  */
	read = M214_CAL_READ_MASK | address;
	checkErr(m214_WriteRegister (handle, M214_CALIBRATION_ADR_REG, read));
	error = m214_WaitForCalibrationAccess(handle, M214_CAL_READ_MASK);
	if(error == 0)
		checkErr(m214_ReadRegister (handle, M214_CALIBRATION_DATA_REG, data));

Error:
	return error;
}



/**************************************************************************************
* WriteCalibrationValue
*
* writes the specified data in the calibration data register
* to the specified calibration address
****************************************************************************************/
Int32 m214_WriteCalibrationValue(M214_HANDLE handle, UInt16 address, UInt16 data) 			   // BS
{
	Int32 error = M214_SUCCESS;
	UInt16 write;

/*  Check Parameters  */
	checkParamRange(address, 0, 0xFFFF, 2);

/*  Perform the write  */	
	checkErr(m214_WriteRegister (handle, M214_CALIBRATION_DATA_REG, data));
	write = M214_CAL_WRITE_MASK | address;
	checkErr(m214_WriteRegister (handle, M214_CALIBRATION_ADR_REG, write));
	error = m214_WaitForCalibrationAccess(handle, M214_CAL_WRITE_MASK);

Error:
	return error;
}


/*************************************************************************
* QueryCalibrationWriteBit
*
* indicates whether the device is still writing to a 
* calibration address
***************************************************************************/
Int32 m214_QueryCalibrationWriteBit(M214_HANDLE handle, Bool* value) 		     // BS
{
	Int32 error = M214_SUCCESS;
	UInt16 data = 0;
	
/*  Check Parameters  */
	checkParamPointer(value, 2);

/*  Read the register and get the write bit  */
	checkErr(m214_ReadRegister (handle, M214_CALIBRATION_ADR_REG, &data));
	if (data & M214_CAL_WRITE_MASK) 
	{ *value = M214_CALIBRATION_BIT_SET; 
	}
	else 
	{ *value = M214_CALIBRATION_BIT_NOT_SET; 
	}
Error:
	return error;
}


/******************************************************************************
* QueryCalibrationReadBit
*
* indicates whether the device is still reading from a
* calibration address
********************************************************************************/
Int32 m214_QueryCalibrationReadBit(M214_HANDLE handle, Bool* value) 			      // BS
{
	Int32 error = M214_SUCCESS;
	UInt16 data = 0;

/*  Check Parameters  */
	checkParamPointer(value, 2);

/*  Read the register and get the read bit  */
	checkErr(m214_ReadRegister (handle, M214_CALIBRATION_ADR_REG, &data));
	if (data & M214_CAL_READ_MASK) 
	{ *value = M214_CALIBRATION_BIT_SET; 
	}
	else 
	{ *value = M214_CALIBRATION_BIT_NOT_SET; 
	}
Error:
	return error;
}



/*******************************************************
* SetVoltage
*
* sets the voltage of the specified output source. the
* range is determined in this function, the output enable
* bit is specified
*
********************************************************/
Int32 m214_SetVoltage(M214_HANDLE handle, UInt8 channel, Real64 voltage) 
{	
	Int32 error = M214_SUCCESS;
	Int32 voltage_bits;
	UInt16 data;
	UInt8 range;
	Real64 fullscale;
	Real64 vperbit;
	
/*  Check Parameters  */
	checkParamRange(channel, 0, 0x3, 2);
	checkParamRange(voltage, -10.0, 10.0, 3);

	if (voltage == 0) 
	{ data = 0;
//		range = 0;
		range = 1;	// Rev 2.1 Set Range to 1 (99mV range) to match offset calibration method
		voltage_bits = 0;
	}
	
	else if (voltage < 0) 
	{ 
		if (voltage >= (M214_NEG_mV_RANGE + 0.001)) { fullscale = M214_RANGE1_FULLSCALE_uV; range = 1; }
		else if (voltage >= (M214_NEG_2V_RANGE + 0.02)) { fullscale =  M214_RANGE2_FULLSCALE_uV; range = 2; }
		else { fullscale =  M214_RANGE3_FULLSCALE_uV; range = 3; }
		vperbit = (Real64)fullscale;
		vperbit = vperbit  / (Real64)0x100000 * (Real64)(1e-6);
		voltage_bits = (Int32)(voltage / vperbit);
	}
	
	else 
	{ if (voltage <= (M214_POS_mV_RANGE - 0.001)) { fullscale = M214_RANGE1_FULLSCALE_uV; range = 1; }
		else if (voltage <= (M214_POS_2V_RANGE - 0.02)) { fullscale = M214_RANGE2_FULLSCALE_uV; range = 2; }
		else { fullscale = M214_RANGE3_FULLSCALE_uV; range = 3; }
		vperbit = (Real64)fullscale;
		vperbit = vperbit  / (Real64)0x100000 * (Real64)(1e-6);
		voltage_bits = (Int32)(voltage / vperbit);
		if (voltage == 10.0) {
			voltage_bits = voltage_bits - 1;	// Rev 3.2 special case to correct for rounding errors
		}
	}
	
	if(handle->fw_rev < 2.0)
	{
		/*  Read/Modify/Write the Voltage Control/Status Register so that we can set the range  */	
		checkErr(m214_ReadRegister (handle, M214_VOLT_CTRL_STAT_REG, &data));
		if (channel == 0) { data &= 0xFFFC; }
		else if (channel == 1) { data &= 0xFFF3; }
		else if (channel == 2) { data &= 0xFFCF; }
		else { data &= 0xFF3F; }
		data |= range << (channel * 2);
		checkErr(m214_WriteRegister (handle, M214_VOLT_CTRL_STAT_REG, data));

		data = (UInt16) voltage_bits & 0x0FFFF;
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_LSB_REG + (4*channel), data));
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		data = data  & 0x8000;		 	// leave TC bit as is
		data = data | ((voltage_bits >> 16) & 0x000F);
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), data));
	}
	else
	{
		data = (UInt16) voltage_bits & 0x0FFFF;
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_LSB_REG + (4*channel), data));
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		data = data  & 0x8100;		 	// leave TC & OE, bits as is and insert range bits
		data = data | ((voltage_bits >> 16) & 0x000F);
		data = data | (range << 4);
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), data));
	}
	
Error:
	return error;
}


/*******************************************************
* GetVoltage
*
* reads the voltage, range, and output enable bit from 
* the desired output source
********************************************************/
Int32 m214_GetVoltage(M214_HANDLE handle, UInt8 channel, Real64* voltage) 
{
	Int32 error = M214_SUCCESS;
	Int32 voltage_bits;
	Real64 fullscale;
	Real64 vperbit;
	UInt16 data;
	UInt8 range;

/*  Check Parameters  */
	checkParamRange(channel, 0, 0x3, 2);
	checkParamPointer(voltage, 3);

	if(handle->fw_rev < 2.0)
	{
		/*  Read Ctrl Stat to get the range  */
		checkErr(m214_ReadRegister (handle, M214_VOLT_CTRL_STAT_REG, &data));
		range = (data >> (channel*2)) & 0x3;
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		voltage_bits = (data << 16) & 0xF0000;
	}
	else
	{
		/*  Read the range and high bits of the voltage value  */
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		voltage_bits = (data << 16) & 0xF0000;
		range = (data >> 4) & 0x3;
	}
	
/*  Read the low bits of the voltage value  */
	checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_LSB_REG + (4*channel), &data));
	voltage_bits |= data;

	if (voltage_bits & 0x80000) { voltage_bits |= 0xFFF80000; }	// if negative then sign extend

/*  Calculate the real voltage from the voltage value  */
	if (range == 0) { fullscale = 0; }
	else if (range == 1) { fullscale = M214_RANGE1_FULLSCALE_uV; }
	else if (range == 2) { fullscale = M214_RANGE2_FULLSCALE_uV; }
	else { fullscale = M214_RANGE3_FULLSCALE_uV; }

	vperbit = (Real64)fullscale;
	vperbit = vperbit  / (Real64)0x100000 * (Real64)(1e-6);
	*voltage = voltage_bits * vperbit;
	
Error:
	return error;
}


/***********************************************************************************************
* GetOutputEnable
*
* checks whether the output is enabled for the specified
* source
************************************************************************************************/
Int32 m214_GetOutputEnable(M214_HANDLE handle, UInt8 channel, Bool *enable) 
{
	Int32 error = M214_SUCCESS;
	UInt16 data = 0;

/*  Check Parameters  */
	checkParamRange(channel, 0, 0x3, 2);
	checkParamPointer(enable, 3);

	if(	handle->fw_rev < 2.0)
	{
		/*  Read the Ctrl/Stat register to get the enable bit  */	
		checkErr(m214_ReadRegister (handle, M214_VOLT_CTRL_STAT_REG, &data));
		*enable = (data >> (M214_OUTPUT_ENABLE_0_BIT_NUM + channel)) & 1;
	}
	else
	{
		/*  Read the Ctrl/Stat register to get the enable bit  */	
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		*enable = (data >> M214_OUTPUT_ENABLE_BIT_NUM) & 1;
	}
	
Error:
	return error;
}


/**********************************************************************************************
* SetOutputEnable
*
* sets output enable bit of specified source to high or low
***********************************************************************************************/
Int32 m214_SetOutputEnable(M214_HANDLE handle, UInt8 channel, Bool enable) 
{
	Int32 error = M214_SUCCESS;
	UInt16 data = 0;

/*  Check Parameters  */	
	checkParamRange(channel, 0, 3, 2);
	checkParamRange(enable, 0, 1, 3);

	if(handle->fw_rev < 2.0)
	{
		/*  Read the register to get the current value  */	
		checkErr(m214_ReadRegister(handle, M214_VOLT_CTRL_STAT_REG, &data));
	
		/*  Clear the output enable bit for the specified channel  */
		if(enable == M214_TRUE)
			data |= 0x1 << (M214_OUTPUT_ENABLE_0_BIT_NUM + channel);
		else
			data &= ~(0x1 << (M214_OUTPUT_ENABLE_0_BIT_NUM + channel));

		/*  Write the new value to the register  */
		checkErr(m214_WriteRegister(handle, M214_VOLT_CTRL_STAT_REG, data));
	}
	else
	{
		/*  Read the register to get the current value  */	
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
	
		/*  Clear the output enable bit for the specified channel  */
		if(enable == M214_TRUE)
			data |= (1 << M214_OUTPUT_ENABLE_BIT_NUM);
		else
			data &= ~(1 << M214_OUTPUT_ENABLE_BIT_NUM);

		/*  Write the new value to the register  */
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), data));
	}
	
Error:
	return error;
}


/*******************************************************
* m214_SetRange
*
* Note: This function is not for general use.
* It should only be used when a manual setting of range
* is required.
********************************************************/
Int32 m214_SetRange (M214_HANDLE handle, UInt8 channel, UInt8 range)
{	Int32 error = M214_SUCCESS;
	UInt16 data;
	Int32 voltage_bits;

/*  Check Parameters  */	
	checkParamRange(channel, 0, 3, 2);
	checkParamRange(range, 0, 3, 3);

	if(	handle->fw_rev < 2.0)
	{
		checkErr(m214_ReadRegister(handle, M214_VOLT_CTRL_STAT_REG, &data));
	
		/*  Clear the range bits for the specified channel  */
		data &= ~(0x3 << (channel * 2));

		/*  Set the range for the specified channel  */	
		data |= range << (channel * 2);

		/*  Write the new value to the register  */
		checkErr(m214_WriteRegister(handle, M214_VOLT_CTRL_STAT_REG, data));
	}
	else
	{
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
	
		/*  Clear the range bits for the specified channel  */
		data &= ~(0x3 << 4);

		/*  Set the range for the specified channel  */	
		data |= (range << 4);

		/*  Write the new value to the register  */
		checkErr(m214_WriteRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), data));
	}

Error:
	return error;
}


/*******************************************************
* m214_GetRange
*
* Note: This function is not for general use.
* It should only be used when a manual setting of range
* is required.
********************************************************/
Int32 m214_GetRange (M214_HANDLE handle, UInt8 channel, UInt8 *range)
{	Int32 error = M214_SUCCESS;
	UInt16 data;

/*  Check Parameters  */	
	checkParamRange(channel, 0, 3, 2);
	checkParamPointer(range, 3);

	if(	handle->fw_rev < 2.0)
	{
		checkErr(m214_ReadRegister(handle, M214_VOLT_CTRL_STAT_REG, &data));
		*range = (data >> (channel * 2)) & 0x3;
	}
	else
	{
		checkErr(m214_ReadRegister (handle, M214_VOLT_SRC0_MSB_REG + (4*channel), &data));
		*range = (data >> 4) & 0x3;
	}

Error:
	return error;
}


/*****************************************************************************
*	m214_ErrorMessage
*
*	Returns a text description of the specified status code.  This includes
*	all VISA errors as well as driver defined errors.
******************************************************************************/
Int32 m214_ErrorMessage (Int32 statusCode, Int8 errorMessage[])
{	Int32 error = M214_SUCCESS;
	
/*  Check parameters  */
	checkParamPointer(errorMessage, 3);

/*  Check driver defined error codes  */
	if((statusCode >= M214_SPECIFIC_ERROR_BASE) && (statusCode < M214_SPECIFIC_ERROR_END))
	{	strcpy(errorMessage, errorMessageTable[statusCode - M214_SPECIFIC_ERROR_BASE]);
	}
	else if((statusCode >= M214_SPECIFIC_WARNING_BASE) && (statusCode < M214_SPECIFIC_WARNING_END))
	{	strcpy(errorMessage, warningMessageTable[statusCode - M214_SPECIFIC_WARNING_BASE]);
	}
	else if (statusCode == 0)
		sprintf(errorMessage, "No Error Occured");
	else
		sprintf(errorMessage, "Unknown Error Code [0x%x]", statusCode);
		
Error:
	return error;
}


/*****************************************************************************
*	m214_GetIdentWord
*
*	Reads the specified offset from the module's ID space and returns
*	the result.
******************************************************************************/
Int32 m214_GetIdentWord (M214_HANDLE handle, UInt16 offset, UInt16 *value)
{	
	Int32 error = M214_SUCCESS;
	UInt8 data;

/*  Check parameters  */
	checkParamPointer(value, 3);

	*value = 0;
	
/*  Send the read command to the serial EEPROM  */	
	checkErr(m214_IDWriteByte(handle, 0x80 | offset));
	
/*  Read upper byte  */	
	checkErr(m214_IDReadByte(handle, &data));
	*value = data << 8;
	
/*  Read lower byte  */	
	checkErr(m214_IDReadByte(handle, &data));
	*value |= data;
	
Error:	
	m214_WriteRegister(handle, M214_ID_ACCESS_REGISTER, 0);
	return error;
}


/*****************************************************************************
*	Modify the functions below this point to use the M214 driver in an 
*	environment not supported by an APIS platform support module. All functions
*	above this point should not have to be modified.
*
*	In general, any call to an APIS function (apis_func()) will need to be
*	modified to communicate with the instrument or operating system being used.
*
******************************************************************************/

/*****************************************************************************
*	m214_Init
*
*	Opens a M214 session to the instrument.  If IDQuery is true then this 
*	function verifies that the instrument is a m214.  If reset is true then
*	this function resets the instrument to it's default state.
******************************************************************************/
Int32 m214_Init (Int32 instr, Bool IDQuery, Bool reset, M214_HANDLE *handle)
{	
	Int32 error = M214_SUCCESS;
	Int32 warning = M214_SUCCESS;
  UInt16 value = 0;
	UInt8 fw_rev = 0;
	UInt8 hw_rev = 0;


/*  Check parameters  */
	checkParamRange(IDQuery, 0, 1, 2);
	checkParamRange(reset, 0, 1, 3);
	checkParamPointer(handle, 4);

/*  Allocate the data structure  */
	(*handle) = 0;
	(*handle) = malloc(sizeof(M214_INSTR));
	if((*handle) == NULL)
	{	error = M214_ERROR_OUT_OF_MEMORY;
		goto Error;
	}
	
	(*handle)->apis_handle = 0;
	apisCheckErrInit(apis_open((APIS_PATH)instr, &(*handle)->apis_handle));

/*  Initialize Attributes  */
	checkErr(m214_RevisionQuery(*handle, NULL, &fw_rev, &hw_rev));
	(*handle)->fw_rev = (Real64)((fw_rev>>4) & 0x0F) + ((Real64)(fw_rev & 0x0F) / 10.0);
	(*handle)->hw_rev = (Real64)((hw_rev>>4) & 0x0F) + ((Real64)(hw_rev & 0x0F) / 10.0);

/*	Perform ID Query  */
		 if(IDQuery == M214_TRUE)
		{	
			checkErr(m214_GetIdentWord (*handle, 17, &value));
			value &= 0x0FFF;
			if (value != (UInt16)M214_VXI_MANF_ID) 
			{	error = M214_ERROR_FAIL_ID_QUERY;
	 			goto Error;
		  }

			checkErr(m214_GetIdentWord (*handle, 18, &value));
			value &= 0x0FFF;
			if (value != (UInt16)M214_VXI_MODEL_CODE) 
			{	error = M214_ERROR_FAIL_ID_QUERY;
				goto Error;
			} 
		}

/* Reset Instrument */
  if (reset == M214_TRUE) 
  {	checkErr(m214_Reset (*handle));
  }


Error:
/*  Do not free handle here.  The Calling function must call m213_Close upon receiving an error  */	
	
	if((error == M214_SUCCESS) && (warning != M214_SUCCESS))	// if no errors then return the warnings			
		error = warning;

	return error;
}


/*****************************************************************************
*	m214_Close
*
*	Close the session to the instrument.
******************************************************************************/
Int32 m214_Close (M214_HANDLE handle)
{	Int32 error = M214_SUCCESS;
	
	checkParamPointer(handle, 1);
	
	if(handle->apis_handle != NULL)
		apisCheckErr(apis_close(handle->apis_handle));
	free(handle);
	
Error:
	return error;
}


/*****************************************************************************
*	m214_Reset
*	
*	Reset the instrument to it's default state
******************************************************************************/
Int32 m214_Reset (M214_HANDLE handle) 
{		
	Int32 error = M214_SUCCESS;
	
	apisCheckErr(apis_reset(handle->apis_handle));

Error:
	return error;
}


/*****************************************************************************
*	m214_ReadRegister
*
*	Read the specified offset from the module's I/O space and returns
*	the result.
******************************************************************************/
Int32 m214_ReadRegister (M214_HANDLE handle, UInt16 offset, UInt16 *value)
{	
	Int32 error = M214_SUCCESS;
/*  Check parameters  */
	checkParamPointer(value, 3);	
	
	apisCheckErr(apis_read(handle->apis_handle, 2, offset, value));
	
Error:
	return error;
}


/*****************************************************************************
*	m214_WriteRegister
*
*	Writes to the specified offset in the module's I/O space
******************************************************************************/
Int32 m214_WriteRegister (M214_HANDLE handle, UInt16 offset, UInt16 value)
{	
	Int32 error = M214_SUCCESS;
	
	apisCheckErr(apis_write(handle->apis_handle, 2, offset, value)); 
	
Error:
	return error;
}


/*****************************************************************************
*	m214_GetApisError
*
*	Returns the saved apis error code
******************************************************************************/
Int32 m214_GetApisError (M214_HANDLE handle, UInt32 *apis_error)
{	
	Int32 error = M214_SUCCESS;
	
	*apis_error = handle->apis_error;
	
	return error;
}


/***********************************************************
* WaitForCalibrationAccess
*
* waits for a calibration address read or write to complete
************************************************************/
Int32 m214_WaitForCalibrationAccess(M214_HANDLE handle, UInt16 mask) 		    
{
	Int32 error = M214_SUCCESS;
	UInt16 data;
	Real64 start_time = 0;
	
	checkErr(m214_ReadRegister (handle, M214_CALIBRATION_ADR_REG, &data));
	start_time = (Real64)apis_tick()/(Real64)APIS_TICKS_PER_SEC;
	while (((((Real64)apis_tick()/(Real64)APIS_TICKS_PER_SEC) - start_time) < M214_CAL_WAIT_TIMEOUT_SEC) 
						&& ((data & mask) != 0)) 
	{	checkErr(m214_ReadRegister (handle, M214_CALIBRATION_ADR_REG, &data));
	}
	if ((data & mask) != 0) 
	{ error = M214_ERROR_CALVAL_ACCESS_TIMEOUT; 
	}
	
Error:
	return error;
}

